import json
import os
import re
import shutil
import time

from datetime import datetime

from COMPS.Data.Simulation import SimulationState
from simtools.DataAccess.DataStore import DataStore
from simtools.ExperimentManager.BaseExperimentManager import BaseExperimentManager
from simtools.SetupParser import SetupParser
from simtools.SimulationCreator.ClusterSimulationsCreator import ClusterSimulationCreator
from simtools.Utilities.Experiments import validate_exp_name
from simtools.Utilities.General import init_logging
from simtools.ExperimentManager import slurm

logger = init_logging("ClusterExperimentManager")

SCRIPT_TEMPLATE = """#!/bin/bash
#SBATCH --job-name {job_name}  # A single job name for the array
#SBATCH --partition={partition} # Which partition to use
#SBATCH --mem-per-cpu={memory} # Memory per core
#SBATCH --time={time_limit}    # Time limit
#SBATCH --ntasks=1
#SBATCH --cpus-per-task={cpu_per_task}
#SBATCH --nodes={nodes}        # How many nodes to request.
{account_option}
#SBATCH --mail-type=ALL 
#SBATCH --mail-user={email}
#SBATCH --array=0-{folders_length} 
#SBATCH -o {working_directory}/slurm-%A_%a.out # STDOUT
#SBATCH -e {working_directory}/slurm-%A_%a.err # STDERR
set -e
folders=({folders})
current_folder=${{folders[$SLURM_ARRAY_TASK_ID]}}
cd {working_directory}/$current_folder
{chmod_command}
{exe_filename} -C ./config.json -I "../Assets;."
{post_process_command}
"""

# TODO: shouldn't have to change the perms on the exe... not sure why it seems to come across as 500

# items that can be on or off, depending on SetupParser configuration
POST_PROCESS_COMMAND = 'python ../Assets/python/dtk_post_process.py'
ACCOUNT_OPTION = '#SBATCH --account={account}'

# https://wikis.nyu.edu/display/NYUHPC/Slurm+Tutorial
# Requires the following type of section in simtools.ini
# [CLUSTERMODE]
# type = CLUSTER
# # Path where the simulation outputs will be stored
# sim_root = P:\Eradication\simulations
#
# # Path for the model to find the input files
# input_root = P:\Projects\dtk-tools-br\examples\inputs
#
# # Path where a 'reporter_plugins' folder containing the needed DLLs
# dll_root = P:\Projects\dtk-tools-br\examples\inputs\dlls
#
# # Path to the model executable
# exe_path = P:\Projects\dtk-tools-br\examples\inputs\Eradication.exe
#
# # Directory containing dtk_post_process.py, if needed
# python_path =
#
# # Resources request
# nodes = 1
# cpu_per_task = 1
# memory_per_cpu = 8192 # MB of memory
#
# # Which email to send the notifications to
# notification_email = braybaud@intven.com
#
# # NYU partition to use
# partition = cpu_short
#
# # Limit time on this job hrs:min:sec
# time_limit = 10:00:00
#
# # if set to something, jobs will run with the specified account in slurm
# account =
#
#


IDMTOOLS_TO_DTKTOOLS_STATUS_MAPPING = {
    slurm.EntityStatus.RUNNING: SimulationState.Running,
    slurm.EntityStatus.SUCCEEDED: SimulationState.Succeeded,
    slurm.EntityStatus.FAILED: SimulationState.Failed
}


class ClusterExperimentManager(BaseExperimentManager):
    """
    Manages the creation, submission, status, parsing, and analysis
    of cluster experiments
    """
    location = 'CLUSTER'

    def __init__(self, experiment, config_builder):
        self.experiment = experiment
        BaseExperimentManager.__init__(self, experiment, config_builder)
        # load the slurm job id if available
        self.job_id = self.load_job_id_for_experiment(experiment=self.experiment)

    def commission_simulations(self):
        pass

    def create_experiment(self, experiment_name, experiment_id=None, suite_id=None):
        import json

        experiment_name = self.clean_experiment_name(experiment_name)

        # Create a unique id
        experiment_id = re.sub('[ :.-]', '_', str(datetime.now()))
        logger.info("Creating exp_id = " + experiment_id)

        # Create the experiment in the base class
        super(ClusterExperimentManager, self).create_experiment(experiment_name, experiment_id, suite_id)

        # Get the path and create it if needed
        experiment_path = self.experiment.get_path()
        os.makedirs(experiment_path, exist_ok=True)
        os.makedirs(os.path.join(experiment_path, "Assets"), exist_ok=True)

        # write suite information
        suite_info = {'id': suite_id}
        suite_file_path = os.path.join(experiment_path, 'suite.json')
        with open(suite_file_path, 'w') as f:
            json.dump(suite_info, f)

    def refresh_experiment(self):
        # Refresh the experiment, in case something else updated it (shouldn't)
        self.experiment = DataStore.get_experiment(self.experiment.exp_id)
        if self.job_id:
            try:
                logger.debug('Getting status for job id: %s' % self.job_id)
                status = slurm.status_for_job_id(job_id=self.job_id)
                logger.debug('Got status for job id: %s %s' % (self.job_id, status))
                dtktools_status = IDMTOOLS_TO_DTKTOOLS_STATUS_MAPPING[status]
            except AttributeError as e:
                # We were unable to retrieve status from sacct. It is being slow. Assume Running until
                # we detect otherwise
                dtktools_status = SimulationState.Running
            # all Cluster simulations currently have identical status, equal to the experiment status
            for s in self.experiment.simulations:
                s.status = dtktools_status
            DataStore.save_experiment(self.experiment, verbose=False)
        else:
            # a job id hasn't been recorded yet because it hasn't been commissioned yet
            dtktools_status = SimulationState.Created
        return dtktools_status

    @staticmethod
    def create_suite(suite_name):
        suite_id = suite_name + '_' + re.sub('[ :.-]', '_', str(datetime.now()))
        logger.info("Creating suite_id = " + suite_id)
        return suite_id

    def hard_delete(self):
        """
        Delete experiment and output data.
        """
        # Delete local simulation data.
        exp_path = self.experiment.get_path()
        if os.path.exists(exp_path):
            try:
                shutil.rmtree(exp_path)
            except Exception as e:
                print("Could not delete path: {}\nReason: {}".format(exp_path, e))

        # Delete in the DB
        from simtools.DataAccess.DataStore import DataStore
        DataStore.delete_experiment(self.experiment)

    # not currently checking the output of this command; it is currently best-effort.
    def cancel_experiment(self):
        super().cancel_experiment()
        result = slurm.cancel_job(job_id=self.job_id)
        logger.info('Cancelled experiment with slurm batch job id: %s' % self.job_id)

    def kill_simulation(self, simulation):
        pass

    def get_simulation_creator(self, work_queue):
        return ClusterSimulationCreator(config_builder=self.config_builder,
                                        initial_tags=self.exp_builder.tags,
                                        work_queue=work_queue,
                                        experiment=self.experiment,
                                        cache=self.cache)

    def run_simulations(self, config_builder=None, exp_name='test', exp_builder=None,
                        suite_id=None, blocking=False, quiet=False, experiment_tags=None):
        """
        Create an experiment with simulations modified according to the specified experiment builder.
        Commission simulations and cache meta-data to local file.
        :assets: A SimulationAssets object if not None (COMPS-land needed for AssetManager)
        """
        # Check experiment name as early as possible
        if not validate_exp_name(exp_name):
            exit()

        # Store the config_builder if passed
        self.config_builder = config_builder or self.config_builder

        # Get the assets from the config builder
        # We just want to check the input files at this point even though it may change later
        self.assets = self.config_builder.get_assets()

        # Check input files existence
        if not self.config_builder.ignore_missing:
            if not self.validate_input_files(needed_file_paths=self.config_builder.get_input_file_paths(),
                                             message="config.json"):
                exit()
            if not self.validate_input_files(needed_file_paths=self.config_builder.get_dll_paths_for_asset_manager(),
                                             message="emodules_map.json"):
                exit()

        # Set the appropriate command line
        self.commandline = self.config_builder.get_commandline()

        # Set the tags
        self.experiment_tags.update(experiment_tags or {})

        # Create the simulations
        self.create_simulations(exp_name=exp_name, exp_builder=exp_builder, suite_id=suite_id, verbose=not quiet)

        # Dump the meta-data file
        self.dump_metadata()

        # Dump the shell script
        script_path = self.dump_shell_script()

        # execute the shell script
        self.job_id = slurm.submit_job(script_path=script_path)
        for s in self.experiment.simulations:
            s.status = SimulationState.Running

        self.record_job_id_for_experiment(job_id=self.job_id, experiment=self.experiment)

        if blocking:
            self.wait_for_finished(verbose=not quiet)

    @staticmethod
    def record_job_id_for_experiment(job_id, experiment):
        record_file = os.path.join(experiment.get_path(), 'slurm-job.json')
        record_data = {
            'job_id': job_id
        }
        with open(record_file, 'w') as f:
            json.dump(record_data, f)

    @staticmethod
    def load_job_id_for_experiment(experiment):
        try:
            record_file = os.path.join(experiment.get_path(), 'slurm-job.json')
            with open(record_file, 'r') as f:
                record_data = json.load(f)
            job_id = record_data['job_id']
        except:
            job_id = None
        return job_id

    @staticmethod
    def status_finished(state):
        return state in (SimulationState.Succeeded, SimulationState.Failed, SimulationState.Canceled)

    # very similar to BaseExperimentManager version, but different due to not tracking individual simulation states
    def wait_for_finished(self, verbose=False, sleep_time=60):
        timeout = 3600 * 24  # 48 hours timeout
        while True:
            status = self.refresh_experiment()

            if timeout < 0:
                raise Exception("Timeout exhausted for experiment {}".format(self.experiment.exp_id))

            # If we are done, exit the loop after updating the individual db simulations
            if self.status_finished(status):
                break

            # Display if verbose
            if verbose:
                print('Experiment: %s status: %s' % (self.job_id, status))
                print('')

            # Wait before going through the loop again
            time.sleep(sleep_time)
            timeout -= sleep_time

        # SHow status one last time
        if verbose:
            print('Experiment: %s status: %s' % (self.job_id, status))
            print('')

        # Refresh the experiment
        self.refresh_experiment()

    def dump_metadata(self):
        metadata = {
            "experiment": self.experiment.exp_name,
            "original_path": self.experiment.get_path(),
            "experiment_tags": self.experiment.tags,
            "simulations": [
                {
                    "id": s.id,
                    "tags": s.tags
                } for s in self.experiment.simulations
            ]
        }
        import json
        with open(os.path.join(self.experiment.get_path(), "metadata.json"), 'w') as fp:
            json.dump(metadata, fp, indent=4)

    def dump_shell_script(self):
        folders = " ".join(s.id for s in self.experiment.simulations)
        working_directory = os.path.abspath(self.experiment.get_path())
        post_process_command = POST_PROCESS_COMMAND if SetupParser.get('python_path') else ''
        account_option = ACCOUNT_OPTION.format(account=SetupParser.get('account')) if SetupParser.get('account') else ''

        if len(self.experiment.simulations) == 0:
            raise AssertionError("No simulations defined. Cannot batch the experiment")

        format_opts = dict(job_name=self.experiment.id,
                           partition=SetupParser.get("partition"),
                           memory=SetupParser.get("memory_per_cpu"),
                           time_limit=SetupParser.get("time_limit"),
                           cpu_per_task=SetupParser.get("cpu_per_task"),
                           nodes=SetupParser.get("nodes"),
                           email=SetupParser.get("notification_email"),
                           folders_length=len(self.experiment.simulations) - 1,
                           folders=folders,
                           working_directory=working_directory,
                           post_process_command=post_process_command,
                           account_option=account_option)

        exec_filename = SetupParser.get('exe_path')

        # if the item is is singularity based, then use that
        if not exec_filename.startswith('singularity'):
            exec_filename = os.path.basename(exec_filename)
            exec_filename = f'../Assets/{exec_filename}'
            chmod_command = f'chmod 744 ../Assets/{exec_filename}'
        else:
            # format exec string if singularity to allow dynamic entries
            exec_filename = exec_filename.format(**format_opts)
            chmod_command = ''

        format_opts['exe_filename'] = exec_filename
        format_opts['chmod_command'] = chmod_command
        shell_script_path = os.path.join(working_directory, "run_script.sh")
        with open(shell_script_path, 'w') as fp:
            fp.write(SCRIPT_TEMPLATE.format(**format_opts))
        return shell_script_path

    @property
    def run_command(self):
        return 'sbatch %s' % os.path.join(self.experiment.get_path(), 'run_script.sh')

    def succeeded(self):
        return True